"""
Read and write PCL .pcd files in python.
dimatura@cmu.edu, 2013-2018

- TODO better API for wacky operations.
- TODO add a cli for common operations.
- TODO deal properly with padding
- TODO deal properly with multicount fields
- TODO better support for rgb nonsense
"""

import re
import struct
import copy
import numpy as np
import warnings
import lzf

try:
    from StringIO import StringIO, BytesIO
except ImportError:
    from io import BytesIO, StringIO

HAS_SENSOR_MSGS = True
try:
    from sensor_msgs.msg import PointField
    import numpy_pc2  # needs sensor_msgs
except ImportError:
    HAS_SENSOR_MSGS = False

__all__ = ['PointCloud',
           'point_cloud_to_path',
           'point_cloud_to_buffer',
           'point_cloud_to_fileobj',
           'point_cloud_from_path',
           'point_cloud_from_buffer',
           'point_cloud_from_fileobj',
           'make_xyz_point_cloud',
           'make_xyz_rgb_point_cloud',
           'make_xyz_label_point_cloud',
           'save_txt',
           'cat_point_clouds',
           'add_fields',
           'update_field',
           'build_ascii_fmtstr',
           'encode_rgb_for_pcl',
           'decode_rgb_from_pcl',
           'save_point_cloud',
           'save_point_cloud_bin',
           'save_point_cloud_bin_compressed',
           'pcd_type_to_numpy_type',
           'numpy_type_to_pcd_type',
           ]

if HAS_SENSOR_MSGS:
    pc2_pcd_type_mappings = [(PointField.INT8, ('I', 1)),
                             (PointField.UINT8, ('U', 1)),
                             (PointField.INT16, ('I', 2)),
                             (PointField.UINT16, ('U', 2)),
                             (PointField.INT32, ('I', 4)),
                             (PointField.UINT32, ('U', 4)),
                             (PointField.FLOAT32, ('F', 4)),
                             (PointField.FLOAT64, ('F', 8))]
    pc2_type_to_pcd_type = dict(pc2_pcd_type_mappings)
    pcd_type_to_pc2_type = dict((q, p) for (p, q) in pc2_pcd_type_mappings)
    __all__.extend(['pcd_type_to_pc2_type', 'pc2_type_to_pcd_type'])

numpy_pcd_type_mappings = [(np.dtype('float32'), ('F', 4)),
                           (np.dtype('float64'), ('F', 8)),
                           (np.dtype('uint8'), ('U', 1)),
                           (np.dtype('uint16'), ('U', 2)),
                           (np.dtype('uint32'), ('U', 4)),
                           (np.dtype('uint64'), ('U', 8)),
                           (np.dtype('int8'), ('I', 1)),
                           (np.dtype('int16'), ('I', 2)),
                           (np.dtype('int32'), ('I', 4)),
                           (np.dtype('int64'), ('I', 8))]
numpy_type_to_pcd_type = dict(numpy_pcd_type_mappings)
pcd_type_to_numpy_type = dict((q, p) for (p, q) in numpy_pcd_type_mappings)


def parse_header(lines):
    """ Parse header of PCD files.
    """
    metadata = {}
    for ln in lines:
        if ln.startswith('#') or len(ln) < 2:
            continue
        match = re.match('(\w+)\s+([\w\s\.]+)', ln)
        if not match:
            warnings.warn("warning: can't understand line: %s" % ln)
            continue
        key, value = match.group(1).lower(), match.group(2)
        if key == 'version':
            metadata[key] = value
        elif key in ('fields', 'type'):
            metadata[key] = value.split()
        elif key in ('size', 'count'):
            metadata[key] = list(map(int, value.split()))
        elif key in ('width', 'height', 'points'):
            metadata[key] = int(value)
        elif key == 'viewpoint':
            metadata[key] = list(map(float, value.split()))
        elif key == 'data':
            metadata[key] = value.strip().lower()
        # TODO apparently count is not required?
    # add some reasonable defaults
    if 'count' not in metadata:
        metadata['count'] = [1]*len(metadata['fields'])
    if 'viewpoint' not in metadata:
        metadata['viewpoint'] = [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0]
    if 'version' not in metadata:
        metadata['version'] = '.7'
    return metadata


def write_header(metadata, rename_padding=False):
    """ Given metadata as dictionary, return a string header.
    """
    template = """\
VERSION {version}
FIELDS {fields}
SIZE {size}
TYPE {type}
COUNT {count}
WIDTH {width}
HEIGHT {height}
VIEWPOINT {viewpoint}
POINTS {points}
DATA {data}
"""
    str_metadata = metadata.copy()

    if not rename_padding:
        str_metadata['fields'] = ' '.join(metadata['fields'])
    else:
        new_fields = []
        for f in metadata['fields']:
            if f == '_':
                new_fields.append('padding')
            else:
                new_fields.append(f)
        str_metadata['fields'] = ' '.join(new_fields)
    str_metadata['size'] = ' '.join(map(str, metadata['size']))
    str_metadata['type'] = ' '.join(metadata['type'])
    str_metadata['count'] = ' '.join(map(str, metadata['count']))
    str_metadata['width'] = str(metadata['width'])
    str_metadata['height'] = str(metadata['height'])
    str_metadata['viewpoint'] = ' '.join(map(str, metadata['viewpoint']))
    str_metadata['points'] = str(metadata['points'])
    tmpl = template.format(**str_metadata)
    return tmpl


def _metadata_is_consistent(metadata):
    """ Sanity check for metadata. Just some basic checks.
    """
    checks = []
    required = ('version', 'fields', 'size', 'width', 'height', 'points',
                'viewpoint', 'data')
    for f in required:
        if f not in metadata:
            print('%s required' % f)
    checks.append((lambda m: all([k in m for k in required]),
                   'missing field'))
    checks.append((lambda m: len(m['type']) == len(m['count']) ==
                   len(m['fields']),
                   'length of type, count and fields must be equal'))
    checks.append((lambda m: m['height'] > 0,
                   'height must be greater than 0'))
    checks.append((lambda m: m['width'] > 0,
                   'width must be greater than 0'))
    checks.append((lambda m: m['points'] > 0,
                   'points must be greater than 0'))
    checks.append((lambda m: m['data'].lower() in ('ascii', 'binary',
                   'binary_compressed'),
                   'unknown data type:'
                   'should be ascii/binary/binary_compressed'))
    ok = True
    for check, msg in checks:
        if not check(metadata):
            print('error:', msg)
            ok = False
    return ok

# def pcd_type_to_numpy(pcd_type, pcd_sz):
#     """ convert from a pcd type string and size to numpy dtype."""
#     typedict = {'F' : { 4:np.float32, 8:np.float64 },
#                 'I' : { 1:np.int8, 2:np.int16, 4:np.int32, 8:np.int64 },
#                 'U' : { 1:np.uint8, 2:np.uint16, 4:np.uint32 , 8:np.uint64 }}
#     return typedict[pcd_type][pcd_sz]


def _build_dtype(metadata):
    """ Build numpy structured array dtype from pcl metadata.

    Note that fields with count > 1 are 'flattened' by creating multiple
    single-count fields.

    *TODO* allow 'proper' multi-count fields.
    """
    fieldnames = []
    typenames = []
    for f, c, t, s in zip(metadata['fields'],
                          metadata['count'],
                          metadata['type'],
                          metadata['size']):
        np_type = pcd_type_to_numpy_type[(t, s)]
        if c == 1:
            fieldnames.append(f)
            typenames.append(np_type)
        else:
            fieldnames.extend(['%s_%04d' % (f, i) for i in range(c)])
            typenames.extend([np_type]*c)
    dtype = np.dtype(list(zip(fieldnames, typenames)))
    return dtype


def build_ascii_fmtstr(pc):
    """ Make a format string for printing to ascii.

    Note %.8f is minimum for rgb.
    """
    fmtstr = []
    for t, cnt in zip(pc.type, pc.count):
        if t == 'F':
            fmtstr.extend(['%.10e']*cnt)
        elif t == 'I':
            fmtstr.extend(['%d']*cnt)
        elif t == 'U':
            fmtstr.extend(['%u']*cnt)
        else:
            raise ValueError("don't know about type %s" % t)
    return fmtstr


def parse_ascii_pc_data(f, dtype, metadata):
    """ Use numpy to parse ascii pointcloud data.
    """
    return np.loadtxt(f, dtype=dtype, delimiter=' ')


def parse_binary_pc_data(f, dtype, metadata):
    rowstep = metadata['points']*dtype.itemsize
    # for some reason pcl adds empty space at the end of files
    buf = f.read(rowstep)
    return np.fromstring(buf, dtype=dtype)


def parse_binary_compressed_pc_data(f, dtype, metadata):
    """ Parse lzf-compressed data.
    Format is undocumented but seems to be:
    - compressed size of data (uint32)
    - uncompressed size of data (uint32)
    - compressed data
    - junk
    """
    fmt = 'II'
    compressed_size, uncompressed_size =\
        struct.unpack(fmt, f.read(struct.calcsize(fmt)))
    compressed_data = f.read(compressed_size)
    # TODO what to use as second argument? if buf is None
    # (compressed > uncompressed)
    # should we read buf as raw binary?
    buf = lzf.decompress(compressed_data, uncompressed_size)
    if len(buf) != uncompressed_size:
        raise IOError('Error decompressing data')
    # the data is stored field-by-field
    pc_data = np.zeros(metadata['width'], dtype=dtype)
    ix = 0
    for dti in range(len(dtype)):
        dt = dtype[dti]
        bytes = dt.itemsize * metadata['width']
        column = np.fromstring(buf[ix:(ix+bytes)], dt)
        pc_data[dtype.names[dti]] = column
        ix += bytes
    return pc_data


def point_cloud_from_fileobj(f):
    """ Parse pointcloud coming from file object f
    """
    header = []
    try:
        while True:
            ln = f.readline().strip()
            if not isinstance(ln, str):
                ln = ln.decode('utf-8')
            header.append(ln)
            if ln.startswith('DATA'):
                metadata = parse_header(header)
                dtype = _build_dtype(metadata)
                break
    except StopIteration:
        raise ValueError("Could not parse header")
    if metadata['data'] == 'ascii':
        pc_data = parse_ascii_pc_data(f, dtype, metadata)
    elif metadata['data'] == 'binary':
        pc_data = parse_binary_pc_data(f, dtype, metadata)
    elif metadata['data'] == 'binary_compressed':
        pc_data = parse_binary_compressed_pc_data(f, dtype, metadata)
    else:
        print('DATA field is neither "ascii" or "binary" or\
                "binary_compressed"')
    return PointCloud(metadata, pc_data)


def point_cloud_from_path(fname):
    """ load point cloud in binary format
    """
    with open(fname, 'rb') as f:
        pc = point_cloud_from_fileobj(f)
    return pc


def point_cloud_from_buffer(buf):
    fileobj = BytesIO(buf)
    pc = point_cloud_from_fileobj(fileobj)
    fileobj.close()  # necessary?
    return pc


def point_cloud_from_string(string):
    with StringIO(string) as fileobj:
        pc = point_cloud_from_fileobj(fileobj)
    return pc


def point_cloud_to_fileobj(pc, fileobj, data_compression=None):
    """ Write pointcloud as .pcd to fileobj.
    If data_compression is not None it overrides pc.data.
    """
    metadata = pc.get_metadata()
    if data_compression is not None:
        data_compression = data_compression.lower()
        assert(data_compression in ('ascii', 'binary', 'binary_compressed'))
        metadata['data'] = data_compression

    header = write_header(metadata).encode('utf-8')
    fileobj.write(header)
    if metadata['data'].lower() == 'ascii':
        fmtstr = build_ascii_fmtstr(pc)
        if pc.pc_data.size == 1:
            np.savetxt(fileobj, pc.pc_data.reshape((1,)), fmt=fmtstr)
        else:
            np.savetxt(fileobj, pc.pc_data, fmt=fmtstr)

    elif metadata['data'].lower() == 'binary':
        fileobj.write(pc.pc_data.tostring())
    elif metadata['data'].lower() == 'binary_compressed':
        # TODO
        # a '_' field is ignored by pcl and breaks compressed point clouds.
        # changing '_' to '_padding' or other name fixes this.
        # admittedly padding shouldn't be compressed in the first place.
        # reorder to column-by-column
        uncompressed_lst = []
        for fieldname in pc.pc_data.dtype.names:
            column = np.ascontiguousarray(pc.pc_data[fieldname]).tostring()
            uncompressed_lst.append(column)
        uncompressed = b''.join(uncompressed_lst)
        uncompressed_size = len(uncompressed)
        # print("uncompressed_size = %r"%(uncompressed_size))
        buf = lzf.compress(uncompressed)
        if buf is None:
            # compression didn't shrink the file
            # TODO what do to do in this case when reading?
            buf = uncompressed
            compressed_size = uncompressed_size
        else:
            compressed_size = len(buf)
        fmt = 'II'
        fileobj.write(struct.pack(fmt, compressed_size, uncompressed_size))
        fileobj.write(buf)
    else:
        raise ValueError('unknown DATA type')
    # we can't close because if it's stringio buf then we can't get value after


def point_cloud_to_path(pc, fname):
    with open(fname, 'w') as f:
        point_cloud_to_fileobj(pc, f)


def point_cloud_to_buffer(pc, data_compression=None):
    fileobj = BytesIO()
    point_cloud_to_fileobj(pc, fileobj, data_compression)
    return fileobj.getvalue()


def save_point_cloud(pc, fname):
    """ Save pointcloud to fname in ascii format.
    """
    with open(fname, 'wb') as f:
        point_cloud_to_fileobj(pc, f, 'ascii')


def save_point_cloud_bin(pc, fname):
    """ Save pointcloud to fname in binary format.
    """
    with open(fname, 'wb') as f:
        point_cloud_to_fileobj(pc, f, 'binary')


def save_point_cloud_bin_compressed(pc, fname):
    """ Save pointcloud to fname in binary compressed format.
    """
    with open(fname, 'wb') as f:
        point_cloud_to_fileobj(pc, f, 'binary_compressed')


def save_xyz_label(pc, fname, use_default_lbl=False):
    """ Save a simple (x y z label) pointcloud, ignoring all other features.
    Label is initialized to 1000, for an obscure program I use.
    """
    md = pc.get_metadata()
    if not use_default_lbl and ('label' not in md['fields']):
        raise Exception('label is not a field in this point cloud')
    with open(fname, 'w') as f:
        for i in xrange(pc.points):
            x, y, z = ['%.4f' % d for d in (
                pc.pc_data['x'][i], pc.pc_data['y'][i], pc.pc_data['z'][i]
                )]
            lbl = '1000' if use_default_lbl else pc.pc_data['label'][i]
            f.write(' '.join((x, y, z, lbl))+'\n')


def save_xyz_intensity_label(pc, fname, use_default_lbl=False):
    """ Save XYZI point cloud.
    """
    md = pc.get_metadata()
    if not use_default_lbl and ('label' not in md['fields']):
        raise Exception('label is not a field in this point cloud')
    if 'intensity' not in md['fields']:
        raise Exception('intensity is not a field in this point cloud')
    with open(fname, 'w') as f:
        for i in xrange(pc.points):
            x, y, z = ['%.4f' % d for d in (
                pc.pc_data['x'][i], pc.pc_data['y'][i], pc.pc_data['z'][i]
                )]
            intensity = '%.4f' % pc.pc_data['intensity'][i]
            lbl = '1000' if use_default_lbl else pc.pc_data['label'][i]
            f.write(' '.join((x, y, z, intensity, lbl))+'\n')


def save_txt(pc, fname, header=True):
    """ Save to csv-style text file, separated by spaces.

    TODO:
    - support multi-count fields.
    - other delimiters.
    """
    with open(fname, 'w') as f:
        if header:
            header_lst = []
            for field_name, cnt in zip(pc.fields, pc.count):
                if cnt == 1:
                    header_lst.append(field_name)
                else:
                    for c in xrange(cnt):
                        header_lst.append('%s_%04d' % (field_name, c))
            f.write(' '.join(header_lst)+'\n')
        fmtstr = build_ascii_fmtstr(pc)
        np.savetxt(f, pc.pc_data, fmt=fmtstr)


def update_field(pc, field, pc_data):
    """ Updates field in-place.
    """
    pc.pc_data[field] = pc_data
    return pc


def add_fields(pc, metadata, pc_data):
    """ Builds copy of pointcloud with extra fields.

    Multi-count fields are sketchy, yet again.
    """
    if len(set(metadata['fields']).intersection(set(pc.fields))) > 0:
        raise Exception("Fields with that name exist.")

    if pc.points != pc_data.size:
        raise Exception("Mismatch in number of points.")

    new_metadata = pc.get_metadata()
    new_metadata['fields'].extend(metadata['fields'])
    new_metadata['count'].extend(metadata['count'])
    new_metadata['size'].extend(metadata['size'])
    new_metadata['type'].extend(metadata['type'])

    # parse metadata to add
    # TODO factor this
    fieldnames, typenames = [], []
    for f, c, t, s in zip(metadata['fields'],
                          metadata['count'],
                          metadata['type'],
                          metadata['size']):
        np_type = pcd_type_to_numpy_type[(t, s)]
        if c == 1:
            fieldnames.append(f)
            typenames.append(np_type)
        else:
            fieldnames.extend(['%s_%04d' % (f, i) for i in xrange(c)])
            typenames.extend([np_type]*c)
    dtype = list(zip(fieldnames, typenames))
    # new dtype. could be inferred?
    new_dtype = [(f, pc.pc_data.dtype[f])
                 for f in pc.pc_data.dtype.names] + dtype

    new_data = np.empty(pc.pc_data.size, new_dtype)
    for n in pc.pc_data.dtype.names:
        new_data[n] = pc.pc_data[n]
    for n, n_tmp in zip(fieldnames, pc_data.dtype.names):
        new_data[n] = pc_data[n_tmp]

    # TODO maybe just all the metadata in the dtype.
    # TODO maybe use composite structured arrays for fields with count > 1
    newpc = PointCloud(new_metadata, new_data)
    return newpc


def cat_point_clouds(pc1, pc2):
    """ Concatenate two point clouds into bigger point cloud.
    Point clouds must have same metadata.
    """
    if len(pc1.fields) != len(pc2.fields):
        raise ValueError("Pointclouds must have same fields")
    new_metadata = pc1.get_metadata()
    new_data = np.concatenate((pc1.pc_data, pc2.pc_data))
    # TODO this only makes sense for unstructured pc?
    new_metadata['width'] = pc1.width+pc2.width
    new_metadata['points'] = pc1.points+pc2.points
    pc3 = PointCloud(new_metadata, new_data)
    return pc3


def make_xyz_point_cloud(xyz, metadata=None):
    """ Make a pointcloud object from xyz array.
    xyz array is cast to float32.
    """
    md = {'version': .7,
          'fields': ['x', 'y', 'z'],
          'size': [4, 4, 4],
          'type': ['F', 'F', 'F'],
          'count': [1, 1, 1],
          'width': len(xyz),
          'height': 1,
          'viewpoint': [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
          'points': len(xyz),
          'data': 'binary'}
    if metadata is not None:
        md.update(metadata)
    xyz = xyz.astype(np.float32)
    pc_data = xyz.view(np.dtype([('x', np.float32),
                                 ('y', np.float32),
                                 ('z', np.float32)])).squeeze()
    # pc_data = np.rec.fromarrays([xyz[:,0], xyz[:,1], xyz[:,2]], dtype=dt)
    # data = np.rec.fromarrays([xyz.T], dtype=dt)
    pc = PointCloud(md, pc_data)
    return pc


def make_xyz_rgb_point_cloud(xyz_rgb, metadata=None):
    """ Make a pointcloud object from xyz array.
    xyz array is assumed to be float32.
    rgb is assumed to be encoded as float32 according to pcl conventions.
    """
    md = {'version': .7,
          'fields': ['x', 'y', 'z', 'rgb'],
          'count': [1, 1, 1, 1],
          'width': len(xyz_rgb),
          'height': 1,
          'viewpoint': [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
          'points': len(xyz_rgb),
          'type': ['F', 'F', 'F', 'F'],
          'size': [4, 4, 4, 4],
          'data': 'binary'}
    if xyz_rgb.dtype != np.float32:
        raise ValueError('array must be float32')
    if metadata is not None:
        md.update(metadata)
    pc_data = xyz_rgb.view(np.dtype([('x', np.float32),
                                     ('y', np.float32),
                                     ('z', np.float32),
                                     ('rgb', np.float32)])).squeeze()
    # pc_data = np.rec.fromarrays([xyz[:,0], xyz[:,1], xyz[:,2]], dtype=dt)
    # data = np.rec.fromarrays([xyz.T], dtype=dt)
    pc = PointCloud(md, pc_data)
    return pc


def encode_rgb_for_pcl(rgb):
    """ Encode bit-packed RGB for use with PCL.

    :param rgb: Nx3 uint8 array with RGB values.
    :rtype: Nx1 float32 array with bit-packed RGB, for PCL.
    """
    assert(rgb.dtype == np.uint8)
    assert(rgb.ndim == 2)
    assert(rgb.shape[1] == 3)
    rgb = rgb.astype(np.uint32)
    rgb = np.array((rgb[:, 0] << 16) | (rgb[:, 1] << 8) | (rgb[:, 2] << 0),
                   dtype=np.uint32)
    rgb.dtype = np.float32
    return rgb


def decode_rgb_from_pcl(rgb):
    """ Decode the bit-packed RGBs used by PCL.

    :param rgb: An Nx1 array.
    :rtype: Nx3 uint8 array with one column per color.
    """

    rgb = rgb.copy()
    rgb.dtype = np.uint32
    r = np.asarray((rgb >> 16) & 255, dtype=np.uint8)
    g = np.asarray((rgb >> 8) & 255, dtype=np.uint8)
    b = np.asarray(rgb & 255, dtype=np.uint8)
    rgb_arr = np.zeros((rgb.size, 3), dtype=np.uint8)
    rgb_arr[:, 0] = r
    rgb_arr[:, 1] = g
    rgb_arr[:, 2] = b
    return rgb_arr


def make_xyz_label_point_cloud(xyzl, label_type='f'):
    """ Make XYZL point cloud from numpy array.

    TODO i labels?
    """
    md = {'version': .7,
          'fields': ['x', 'y', 'z', 'label'],
          'count': [1, 1, 1, 1],
          'width': len(xyzl),
          'height': 1,
          'viewpoint': [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
          'points': len(xyzl),
          'data': 'ASCII'}
    if label_type.lower() == 'f':
        md['size'] = [4, 4, 4, 4]
        md['type'] = ['F', 'F', 'F', 'F']
    elif label_type.lower() == 'u':
        md['size'] = [4, 4, 4, 1]
        md['type'] = ['F', 'F', 'F', 'U']
    else:
        raise ValueError('label type must be F or U')
    # TODO use .view()
    xyzl = xyzl.astype(np.float32)
    dt = np.dtype([('x', np.float32), ('y', np.float32), ('z', np.float32),
                   ('label', np.float32)])
    pc_data = np.rec.fromarrays([xyzl[:, 0], xyzl[:, 1], xyzl[:, 2],
                                 xyzl[:, 3]], dtype=dt)
    pc = PointCloud(md, pc_data)
    return pc


class PointCloud(object):
    """ Wrapper for point cloud data.

    The variable members of this class parallel the ones used by
    the PCD metadata (and similar to PCL and ROS PointCloud2 messages),

    ``pc_data`` holds the actual data as a structured numpy array.

    The other relevant metadata variables are:

    - ``version``: Version, usually .7
    - ``fields``: Field names, e.g. ``['x', 'y' 'z']``.
    - ``size.`: Field sizes in bytes, e.g. ``[4, 4, 4]``.
    - ``count``: Counts per field e.g. ``[1, 1, 1]``. NB: Multi-count field
      support is sketchy.
    - ``width``: Number of points, for unstructured point clouds (assumed by
      most operations).
    - ``height``: 1 for unstructured point clouds (again, what we assume most
      of the time.
    - ``viewpoint``: A pose for the viewpoint of the cloud, as
      x y z qw qx qy qz, e.g. ``[0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0]``.
    - ``points``: Number of points.
    - ``type``: Data type of each field, e.g. ``[F, F, F]``.
    - ``data``: Data storage format. One of ``ascii``, ``binary`` or ``binary_compressed``.

    See `PCL docs <http://pointclouds.org/documentation/tutorials/pcd_file_format.php>`__
    for more information.
    """

    def __init__(self, metadata, pc_data):
        self.metadata_keys = metadata.keys()
        self.__dict__.update(metadata)
        self.pc_data = pc_data
        self.check_sanity()

    def get_metadata(self):
        """ returns copy of metadata """
        metadata = {}
        for k in self.metadata_keys:
            metadata[k] = copy.copy(getattr(self, k))
        return metadata

    def check_sanity(self):
        # pdb.set_trace()
        md = self.get_metadata()
        assert(_metadata_is_consistent(md))
        assert(self.pc_data.size == self.points)
        assert(self.width*self.height == self.points)
        assert(len(self.fields) == len(self.count))
        assert(len(self.fields) == len(self.type))

    def save(self, fname):
        self.save_pcd(fname, 'ascii')

    def save_pcd(self, fname, compression=None, **kwargs):
        if 'data_compression' in kwargs:
            warnings.warn('data_compression keyword is deprecated for'
                          ' compression')
            compression = kwargs['data_compression']
        with open(fname, 'wb') as f:
            point_cloud_to_fileobj(self, f, compression)

    def save_pcd_to_fileobj(self, fileobj, compression=None, **kwargs):
        if 'data_compression' in kwargs:
            warnings.warn('data_compression keyword is deprecated for'
                          ' compression')
            compression = kwargs['data_compression']
        point_cloud_to_fileobj(self, fileobj, compression)

    def save_pcd_to_buffer(self, compression=None, **kwargs):
        if 'data_compression' in kwargs:
            warnings.warn('data_compression keyword is deprecated for'
                          ' compression')
            compression = kwargs['data_compression']
        return point_cloud_to_buffer(self, compression)

    def save_txt(self, fname):
        save_txt(self, fname)

    def save_xyz_label(self, fname, **kwargs):
        save_xyz_label(self, fname, **kwargs)

    def save_xyz_intensity_label(self, fname, **kwargs):
        save_xyz_intensity_label(self, fname, **kwargs)

    def copy(self):
        new_pc_data = np.copy(self.pc_data)
        new_metadata = self.get_metadata()
        return PointCloud(new_metadata, new_pc_data)

    def to_msg(self):
        if not HAS_SENSOR_MSGS:
            raise Exception('ROS sensor_msgs not found')
        # TODO is there some metadata we want to attach?
        return numpy_pc2.array_to_pointcloud2(self.pc_data)

    @staticmethod
    def from_path(fname):
        return point_cloud_from_path(fname)

    @staticmethod
    def from_fileobj(fileobj):
        return point_cloud_from_fileobj(fileobj)

    @staticmethod
    def from_string(string):
        return point_cloud_from_string(string)

    @staticmethod
    def from_buffer(buf):
        try:
            return point_cloud_from_buffer(buf)
        except TypeError:
            # The old method could handle strings. The new one should also be able to.
            return PointCloud.from_string(buf)

    @staticmethod
    def from_array(arr):
        """ create a PointCloud object from an array.
        """
        pc_data = arr.copy()
        md = {'version': .7,
              'fields': [],
              'size': [],
              'count': [],
              'width': 0,
              'height': 1,
              'viewpoint': [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
              'points': 0,
              'type': [],
              'data': 'binary_compressed'}
        md['fields'] = pc_data.dtype.names
        for field in md['fields']:
            type_, size_ =\
                numpy_type_to_pcd_type[pc_data.dtype.fields[field][0]]
            md['type'].append(type_)
            md['size'].append(size_)
            # TODO handle multicount
            md['count'].append(1)
        md['width'] = len(pc_data)
        md['points'] = len(pc_data)
        pc = PointCloud(md, pc_data)
        return pc

    @staticmethod
    def from_msg(msg, squeeze=True):
        """ from pointcloud2 msg
        squeeze: fix when clouds get 1 as first dim
        """
        if not HAS_SENSOR_MSGS:
            raise NotImplementedError('ROS sensor_msgs not found')
        md = {'version': .7,
              'fields': [],
              'size': [],
              'count': [],
              'width': msg.width,
              'height': msg.height,
              'viewpoint': [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
              'points': 0,
              'type': [],
              'data': 'binary_compressed'}
        for field in msg.fields:
            md['fields'].append(field.name)
            t, s = pc2_type_to_pcd_type[field.datatype]
            md['type'].append(t)
            md['size'].append(s)
            # TODO handle multicount correctly
            if field.count > 1:
                warnings.warn('fields with count > 1 are not well tested')
            md['count'].append(field.count)
        pc_array = numpy_pc2.pointcloud2_to_array(msg)
        pc_data = pc_array.reshape(-1)
        md['height'], md['width'] = pc_array.shape
        md['points'] = len(pc_data)
        pc = PointCloud(md, pc_data)
        return pc



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protobuf/server.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='protobuf/server.proto',
  package='Server',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x15protobuf/server.proto\x12\x06Server\"\xc2\x03\n\rServerCommand\x12\x0c\n\x04UUID\x18\x01 \x01(\t\x12\x15\n\x0blog_message\x18\x02 \x01(\tH\x00\x12\x18\n\x0e\x63\x61pture_screen\x18\x03 \x01(\x08H\x00\x12\x33\n\x12\x61\x64\x64_custom_control\x18\x04 \x01(\x0b\x32\x15.Server.CustomControlH\x00\x12\'\n\nset_camera\x18\x05 \x01(\x0b\x32\x11.Server.SetCameraH\x00\x12\'\n\nadd_object\x18\x06 \x01(\x0b\x32\x11.Server.AddObjectH\x00\x12;\n\x15set_key_event_handler\x18\x07 \x01(\x0b\x32\x1a.Server.SetKeyEventHandlerH\x00\x12-\n\rremove_object\x18\x08 \x01(\x0b\x32\x14.Server.RemoveObjectH\x00\x12<\n\x15remove_custom_control\x18\t \x01(\x0b\x32\x1b.Server.RemoveCustomControlH\x00\x12\x36\n\x12set_custom_control\x18\n \x01(\x0b\x32\x18.Server.SetCustomControlH\x00\x42\t\n\x07\x43ommand\"\xbf\x06\n\rCustomControl\x12.\n\x06slider\x18\x01 \x01(\x0b\x32\x1c.Server.CustomControl.SliderH\x00\x12\x32\n\x08\x63heckbox\x18\x02 \x01(\x0b\x32\x1e.Server.CustomControl.CheckBoxH\x00\x12\x30\n\x07textbox\x18\x03 \x01(\x0b\x32\x1d.Server.CustomControl.TextBoxH\x00\x12\x34\n\tselectbox\x18\x04 \x01(\x0b\x32\x1f.Server.CustomControl.SelectBoxH\x00\x12.\n\x06\x62utton\x18\x05 \x01(\x0b\x32\x1c.Server.CustomControl.ButtonH\x00\x12\x39\n\x0c\x63olor_picker\x18\x06 \x01(\x0b\x32!.Server.CustomControl.ColorPickerH\x00\x12.\n\x06\x66older\x18\x07 \x01(\x0b\x32\x1c.Server.CustomControl.FolderH\x00\x1a\x62\n\x06Slider\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03min\x18\x02 \x01(\x02\x12\x0b\n\x03max\x18\x03 \x01(\x02\x12\x0c\n\x04step\x18\x04 \x01(\x02\x12\x12\n\ninit_value\x18\x05 \x01(\x02\x12\x0e\n\x06parent\x18\x06 \x01(\t\x1a<\n\x08\x43heckBox\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x12\n\ninit_value\x18\x02 \x01(\x08\x12\x0e\n\x06parent\x18\x03 \x01(\t\x1a;\n\x07TextBox\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x12\n\ninit_value\x18\x02 \x01(\t\x12\x0e\n\x06parent\x18\x03 \x01(\t\x1aL\n\tSelectBox\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05items\x18\x02 \x03(\t\x12\x12\n\ninit_value\x18\x03 \x01(\t\x12\x0e\n\x06parent\x18\x04 \x01(\t\x1a&\n\x06\x42utton\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06parent\x18\x02 \x01(\t\x1a?\n\x0b\x43olorPicker\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x12\n\ninit_value\x18\x02 \x01(\t\x12\x0e\n\x06parent\x18\x03 \x01(\t\x1a&\n\x06\x46older\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06parent\x18\x02 \x01(\tB\t\n\x07\x43ontrol\"A\n\x13RemoveCustomControl\x12\r\n\x03\x61ll\x18\x01 \x01(\x08H\x00\x12\x11\n\x07\x62y_uuid\x18\x02 \x01(\tH\x00\x42\x08\n\x06Object\"\x9f\x01\n\tSetCamera\x12\x19\n\x0fperspective_fov\x18\x01 \x01(\x02H\x00\x12%\n\x1borthographic_frustum_height\x18\x02 \x01(\x02H\x00\x12#\n\x08position\x18\x03 \x01(\x0b\x32\x0f.Server.VecXYZfH\x00\x12!\n\x06target\x18\x04 \x01(\x0b\x32\x0f.Server.VecXYZfH\x00\x42\x08\n\x06\x43\x61mera\"\xb4\x06\n\tAddObject\x12-\n\x08line_set\x18\x01 \x01(\x0b\x32\x19.Server.AddObject.LineSetH\x00\x12\x33\n\x0bpoint_cloud\x18\x02 \x01(\x0b\x32\x1c.Server.AddObject.PointCloudH\x00\x12,\n\x07overlay\x18\x03 \x01(\x0b\x32\x19.Server.AddObject.OverlayH\x00\x12&\n\x04mesh\x18\x04 \x01(\x0b\x32\x16.Server.AddObject.MeshH\x00\x1a\x81\x01\n\x07LineSet\x12\x1f\n\x06points\x18\x01 \x03(\x0b\x32\x0f.Server.VecXYZf\x12\x12\n\nfrom_index\x18\x02 \x03(\x05\x12\x10\n\x08to_index\x18\x03 \x03(\x05\x12\x1f\n\x06\x63olors\x18\x04 \x03(\x0b\x32\x0f.Server.VecRGBf\x12\x0e\n\x06widths\x18\x05 \x03(\x02\x1a\x32\n\nPointCloud\x12\x10\n\x08pcd_data\x18\x01 \x01(\x0c\x12\x12\n\npoint_size\x18\x02 \x01(\x02\x1a\x97\x02\n\x07Overlay\x12!\n\x08position\x18\x01 \x01(\x0b\x32\x0f.Server.VecXYZf\x12\x0e\n\x04html\x18\x02 \x01(\tH\x00\x12\x30\n\x05image\x18\x03 \x01(\x0b\x32\x1f.Server.AddObject.Overlay.ImageH\x00\x12\x36\n\x04type\x18\x04 \x01(\x0e\x32(.Server.AddObject.Overlay.CoordinateType\x1a$\n\x05Image\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\r\n\x05width\x18\x02 \x01(\x05\"=\n\x0e\x43oordinateType\x12\x14\n\x10WORLD_COORDINATE\x10\x00\x12\x15\n\x11SCREEN_COORDINATE\x10\x01\x42\n\n\x08\x43ontents\x1a\x90\x01\n\x04Mesh\x12\x1f\n\x06points\x18\x01 \x03(\x0b\x32\x0f.Server.VecXYZf\x12\x16\n\x0evertex_a_index\x18\x02 \x03(\x05\x12\x16\n\x0evertex_b_index\x18\x03 \x03(\x05\x12\x16\n\x0evertex_c_index\x18\x04 \x03(\x05\x12\x1f\n\x06\x63olors\x18\x05 \x03(\x0b\x32\x0f.Server.VecRGBfB\x08\n\x06Object\":\n\x0cRemoveObject\x12\r\n\x03\x61ll\x18\x01 \x01(\x08H\x00\x12\x11\n\x07\x62y_uuid\x18\x02 \x01(\tH\x00\x42\x08\n\x06Object\"U\n\x12SetKeyEventHandler\x12\x11\n\x07keydown\x18\x01 \x01(\x08H\x00\x12\x0f\n\x05keyup\x18\x02 \x01(\x08H\x00\x12\x12\n\x08keypress\x18\x03 \x01(\x08H\x00\x42\x07\n\x05\x45vent\"*\n\x07VecXYZf\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\"*\n\x07VecRGBf\x12\t\n\x01r\x18\x01 \x01(\x02\x12\t\n\x01g\x18\x02 \x01(\x02\x12\t\n\x01\x62\x18\x03 \x01(\x02\"\xdb\x06\n\x10SetCustomControl\x12\x0e\n\x06target\x18\x01 \x01(\t\x12\x31\n\x06slider\x18\x02 \x01(\x0b\x32\x1f.Server.SetCustomControl.SliderH\x00\x12\x35\n\x08\x63heckbox\x18\x03 \x01(\x0b\x32!.Server.SetCustomControl.CheckBoxH\x00\x12\x33\n\x07textbox\x18\x04 \x01(\x0b\x32 .Server.SetCustomControl.TextBoxH\x00\x12\x37\n\tselectbox\x18\x05 \x01(\x0b\x32\".Server.SetCustomControl.SelectBoxH\x00\x12\x31\n\x06\x62utton\x18\x06 \x01(\x0b\x32\x1f.Server.SetCustomControl.ButtonH\x00\x12<\n\x0c\x63olor_picker\x18\x07 \x01(\x0b\x32$.Server.SetCustomControl.ColorPickerH\x00\x1a\x92\x01\n\x06Slider\x12\x11\n\x04name\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x10\n\x03min\x18\x02 \x01(\x02H\x01\x88\x01\x01\x12\x10\n\x03max\x18\x03 \x01(\x02H\x02\x88\x01\x01\x12\x11\n\x04step\x18\x04 \x01(\x02H\x03\x88\x01\x01\x12\x12\n\x05value\x18\x05 \x01(\x02H\x04\x88\x01\x01\x42\x07\n\x05_nameB\x06\n\x04_minB\x06\n\x04_maxB\x07\n\x05_stepB\x08\n\x06_value\x1a\x44\n\x08\x43heckBox\x12\x11\n\x04name\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x12\n\x05value\x18\x02 \x01(\x08H\x01\x88\x01\x01\x42\x07\n\x05_nameB\x08\n\x06_value\x1a\x43\n\x07TextBox\x12\x11\n\x04name\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x12\n\x05value\x18\x02 \x01(\tH\x01\x88\x01\x01\x42\x07\n\x05_nameB\x08\n\x06_value\x1aT\n\tSelectBox\x12\x11\n\x04name\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\r\n\x05items\x18\x02 \x03(\t\x12\x12\n\x05value\x18\x03 \x01(\tH\x01\x88\x01\x01\x42\x07\n\x05_nameB\x08\n\x06_value\x1a$\n\x06\x42utton\x12\x11\n\x04name\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\x07\n\x05_name\x1aG\n\x0b\x43olorPicker\x12\x11\n\x04name\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x12\n\x05value\x18\x02 \x01(\tH\x01\x88\x01\x01\x42\x07\n\x05_nameB\x08\n\x06_valueB\t\n\x07\x43ontrolb\x06proto3'
)



_ADDOBJECT_OVERLAY_COORDINATETYPE = _descriptor.EnumDescriptor(
  name='CoordinateType',
  full_name='Server.AddObject.Overlay.CoordinateType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='WORLD_COORDINATE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SCREEN_COORDINATE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2140,
  serialized_end=2201,
)
_sym_db.RegisterEnumDescriptor(_ADDOBJECT_OVERLAY_COORDINATETYPE)


_SERVERCOMMAND = _descriptor.Descriptor(
  name='ServerCommand',
  full_name='Server.ServerCommand',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='UUID', full_name='Server.ServerCommand.UUID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='log_message', full_name='Server.ServerCommand.log_message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='capture_screen', full_name='Server.ServerCommand.capture_screen', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='add_custom_control', full_name='Server.ServerCommand.add_custom_control', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='set_camera', full_name='Server.ServerCommand.set_camera', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='add_object', full_name='Server.ServerCommand.add_object', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='set_key_event_handler', full_name='Server.ServerCommand.set_key_event_handler', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='remove_object', full_name='Server.ServerCommand.remove_object', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='remove_custom_control', full_name='Server.ServerCommand.remove_custom_control', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='set_custom_control', full_name='Server.ServerCommand.set_custom_control', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Command', full_name='Server.ServerCommand.Command',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=34,
  serialized_end=484,
)


_CUSTOMCONTROL_SLIDER = _descriptor.Descriptor(
  name='Slider',
  full_name='Server.CustomControl.Slider',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Server.CustomControl.Slider.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min', full_name='Server.CustomControl.Slider.min', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max', full_name='Server.CustomControl.Slider.max', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='step', full_name='Server.CustomControl.Slider.step', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='init_value', full_name='Server.CustomControl.Slider.init_value', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent', full_name='Server.CustomControl.Slider.parent', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=863,
  serialized_end=961,
)

_CUSTOMCONTROL_CHECKBOX = _descriptor.Descriptor(
  name='CheckBox',
  full_name='Server.CustomControl.CheckBox',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Server.CustomControl.CheckBox.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='init_value', full_name='Server.CustomControl.CheckBox.init_value', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent', full_name='Server.CustomControl.CheckBox.parent', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=963,
  serialized_end=1023,
)

_CUSTOMCONTROL_TEXTBOX = _descriptor.Descriptor(
  name='TextBox',
  full_name='Server.CustomControl.TextBox',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Server.CustomControl.TextBox.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='init_value', full_name='Server.CustomControl.TextBox.init_value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent', full_name='Server.CustomControl.TextBox.parent', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1025,
  serialized_end=1084,
)

_CUSTOMCONTROL_SELECTBOX = _descriptor.Descriptor(
  name='SelectBox',
  full_name='Server.CustomControl.SelectBox',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Server.CustomControl.SelectBox.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='items', full_name='Server.CustomControl.SelectBox.items', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='init_value', full_name='Server.CustomControl.SelectBox.init_value', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent', full_name='Server.CustomControl.SelectBox.parent', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1086,
  serialized_end=1162,
)

_CUSTOMCONTROL_BUTTON = _descriptor.Descriptor(
  name='Button',
  full_name='Server.CustomControl.Button',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Server.CustomControl.Button.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent', full_name='Server.CustomControl.Button.parent', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1164,
  serialized_end=1202,
)

_CUSTOMCONTROL_COLORPICKER = _descriptor.Descriptor(
  name='ColorPicker',
  full_name='Server.CustomControl.ColorPicker',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Server.CustomControl.ColorPicker.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='init_value', full_name='Server.CustomControl.ColorPicker.init_value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent', full_name='Server.CustomControl.ColorPicker.parent', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1204,
  serialized_end=1267,
)

_CUSTOMCONTROL_FOLDER = _descriptor.Descriptor(
  name='Folder',
  full_name='Server.CustomControl.Folder',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Server.CustomControl.Folder.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent', full_name='Server.CustomControl.Folder.parent', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1269,
  serialized_end=1307,
)

_CUSTOMCONTROL = _descriptor.Descriptor(
  name='CustomControl',
  full_name='Server.CustomControl',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='slider', full_name='Server.CustomControl.slider', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='checkbox', full_name='Server.CustomControl.checkbox', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='textbox', full_name='Server.CustomControl.textbox', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='selectbox', full_name='Server.CustomControl.selectbox', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='button', full_name='Server.CustomControl.button', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='color_picker', full_name='Server.CustomControl.color_picker', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='folder', full_name='Server.CustomControl.folder', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_CUSTOMCONTROL_SLIDER, _CUSTOMCONTROL_CHECKBOX, _CUSTOMCONTROL_TEXTBOX, _CUSTOMCONTROL_SELECTBOX, _CUSTOMCONTROL_BUTTON, _CUSTOMCONTROL_COLORPICKER, _CUSTOMCONTROL_FOLDER, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Control', full_name='Server.CustomControl.Control',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=487,
  serialized_end=1318,
)


_REMOVECUSTOMCONTROL = _descriptor.Descriptor(
  name='RemoveCustomControl',
  full_name='Server.RemoveCustomControl',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='all', full_name='Server.RemoveCustomControl.all', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='by_uuid', full_name='Server.RemoveCustomControl.by_uuid', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Object', full_name='Server.RemoveCustomControl.Object',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1320,
  serialized_end=1385,
)


_SETCAMERA = _descriptor.Descriptor(
  name='SetCamera',
  full_name='Server.SetCamera',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='perspective_fov', full_name='Server.SetCamera.perspective_fov', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='orthographic_frustum_height', full_name='Server.SetCamera.orthographic_frustum_height', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='position', full_name='Server.SetCamera.position', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target', full_name='Server.SetCamera.target', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Camera', full_name='Server.SetCamera.Camera',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1388,
  serialized_end=1547,
)


_ADDOBJECT_LINESET = _descriptor.Descriptor(
  name='LineSet',
  full_name='Server.AddObject.LineSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='points', full_name='Server.AddObject.LineSet.points', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='from_index', full_name='Server.AddObject.LineSet.from_index', index=1,
      number=2, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='to_index', full_name='Server.AddObject.LineSet.to_index', index=2,
      number=3, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='colors', full_name='Server.AddObject.LineSet.colors', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='widths', full_name='Server.AddObject.LineSet.widths', index=4,
      number=5, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1750,
  serialized_end=1879,
)

_ADDOBJECT_POINTCLOUD = _descriptor.Descriptor(
  name='PointCloud',
  full_name='Server.AddObject.PointCloud',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='pcd_data', full_name='Server.AddObject.PointCloud.pcd_data', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='point_size', full_name='Server.AddObject.PointCloud.point_size', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1881,
  serialized_end=1931,
)

_ADDOBJECT_OVERLAY_IMAGE = _descriptor.Descriptor(
  name='Image',
  full_name='Server.AddObject.Overlay.Image',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data', full_name='Server.AddObject.Overlay.Image.data', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='width', full_name='Server.AddObject.Overlay.Image.width', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2102,
  serialized_end=2138,
)

_ADDOBJECT_OVERLAY = _descriptor.Descriptor(
  name='Overlay',
  full_name='Server.AddObject.Overlay',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='position', full_name='Server.AddObject.Overlay.position', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='html', full_name='Server.AddObject.Overlay.html', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='image', full_name='Server.AddObject.Overlay.image', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='Server.AddObject.Overlay.type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_ADDOBJECT_OVERLAY_IMAGE, ],
  enum_types=[
    _ADDOBJECT_OVERLAY_COORDINATETYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Contents', full_name='Server.AddObject.Overlay.Contents',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1934,
  serialized_end=2213,
)

_ADDOBJECT_MESH = _descriptor.Descriptor(
  name='Mesh',
  full_name='Server.AddObject.Mesh',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='points', full_name='Server.AddObject.Mesh.points', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='vertex_a_index', full_name='Server.AddObject.Mesh.vertex_a_index', index=1,
      number=2, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='vertex_b_index', full_name='Server.AddObject.Mesh.vertex_b_index', index=2,
      number=3, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='vertex_c_index', full_name='Server.AddObject.Mesh.vertex_c_index', index=3,
      number=4, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='colors', full_name='Server.AddObject.Mesh.colors', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2216,
  serialized_end=2360,
)

_ADDOBJECT = _descriptor.Descriptor(
  name='AddObject',
  full_name='Server.AddObject',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='line_set', full_name='Server.AddObject.line_set', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='point_cloud', full_name='Server.AddObject.point_cloud', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='overlay', full_name='Server.AddObject.overlay', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mesh', full_name='Server.AddObject.mesh', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_ADDOBJECT_LINESET, _ADDOBJECT_POINTCLOUD, _ADDOBJECT_OVERLAY, _ADDOBJECT_MESH, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Object', full_name='Server.AddObject.Object',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1550,
  serialized_end=2370,
)


_REMOVEOBJECT = _descriptor.Descriptor(
  name='RemoveObject',
  full_name='Server.RemoveObject',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='all', full_name='Server.RemoveObject.all', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='by_uuid', full_name='Server.RemoveObject.by_uuid', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Object', full_name='Server.RemoveObject.Object',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=2372,
  serialized_end=2430,
)


_SETKEYEVENTHANDLER = _descriptor.Descriptor(
  name='SetKeyEventHandler',
  full_name='Server.SetKeyEventHandler',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='keydown', full_name='Server.SetKeyEventHandler.keydown', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='keyup', full_name='Server.SetKeyEventHandler.keyup', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='keypress', full_name='Server.SetKeyEventHandler.keypress', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Event', full_name='Server.SetKeyEventHandler.Event',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=2432,
  serialized_end=2517,
)


_VECXYZF = _descriptor.Descriptor(
  name='VecXYZf',
  full_name='Server.VecXYZf',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='Server.VecXYZf.x', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y', full_name='Server.VecXYZf.y', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='z', full_name='Server.VecXYZf.z', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2519,
  serialized_end=2561,
)


_VECRGBF = _descriptor.Descriptor(
  name='VecRGBf',
  full_name='Server.VecRGBf',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='r', full_name='Server.VecRGBf.r', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='g', full_name='Server.VecRGBf.g', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='b', full_name='Server.VecRGBf.b', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2563,
  serialized_end=2605,
)


_SETCUSTOMCONTROL_SLIDER = _descriptor.Descriptor(
  name='Slider',
  full_name='Server.SetCustomControl.Slider',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Server.SetCustomControl.Slider.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min', full_name='Server.SetCustomControl.Slider.min', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max', full_name='Server.SetCustomControl.Slider.max', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='step', full_name='Server.SetCustomControl.Slider.step', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='Server.SetCustomControl.Slider.value', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_name', full_name='Server.SetCustomControl.Slider._name',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_min', full_name='Server.SetCustomControl.Slider._min',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_max', full_name='Server.SetCustomControl.Slider._max',
      index=2, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_step', full_name='Server.SetCustomControl.Slider._step',
      index=3, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_value', full_name='Server.SetCustomControl.Slider._value',
      index=4, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=2974,
  serialized_end=3120,
)

_SETCUSTOMCONTROL_CHECKBOX = _descriptor.Descriptor(
  name='CheckBox',
  full_name='Server.SetCustomControl.CheckBox',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Server.SetCustomControl.CheckBox.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='Server.SetCustomControl.CheckBox.value', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_name', full_name='Server.SetCustomControl.CheckBox._name',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_value', full_name='Server.SetCustomControl.CheckBox._value',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=3122,
  serialized_end=3190,
)

_SETCUSTOMCONTROL_TEXTBOX = _descriptor.Descriptor(
  name='TextBox',
  full_name='Server.SetCustomControl.TextBox',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Server.SetCustomControl.TextBox.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='Server.SetCustomControl.TextBox.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_name', full_name='Server.SetCustomControl.TextBox._name',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_value', full_name='Server.SetCustomControl.TextBox._value',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=3192,
  serialized_end=3259,
)

_SETCUSTOMCONTROL_SELECTBOX = _descriptor.Descriptor(
  name='SelectBox',
  full_name='Server.SetCustomControl.SelectBox',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Server.SetCustomControl.SelectBox.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='items', full_name='Server.SetCustomControl.SelectBox.items', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='Server.SetCustomControl.SelectBox.value', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_name', full_name='Server.SetCustomControl.SelectBox._name',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_value', full_name='Server.SetCustomControl.SelectBox._value',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=3261,
  serialized_end=3345,
)

_SETCUSTOMCONTROL_BUTTON = _descriptor.Descriptor(
  name='Button',
  full_name='Server.SetCustomControl.Button',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Server.SetCustomControl.Button.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_name', full_name='Server.SetCustomControl.Button._name',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=3347,
  serialized_end=3383,
)

_SETCUSTOMCONTROL_COLORPICKER = _descriptor.Descriptor(
  name='ColorPicker',
  full_name='Server.SetCustomControl.ColorPicker',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='Server.SetCustomControl.ColorPicker.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='Server.SetCustomControl.ColorPicker.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_name', full_name='Server.SetCustomControl.ColorPicker._name',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_value', full_name='Server.SetCustomControl.ColorPicker._value',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=3385,
  serialized_end=3456,
)

_SETCUSTOMCONTROL = _descriptor.Descriptor(
  name='SetCustomControl',
  full_name='Server.SetCustomControl',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='target', full_name='Server.SetCustomControl.target', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='slider', full_name='Server.SetCustomControl.slider', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='checkbox', full_name='Server.SetCustomControl.checkbox', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='textbox', full_name='Server.SetCustomControl.textbox', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='selectbox', full_name='Server.SetCustomControl.selectbox', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='button', full_name='Server.SetCustomControl.button', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='color_picker', full_name='Server.SetCustomControl.color_picker', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SETCUSTOMCONTROL_SLIDER, _SETCUSTOMCONTROL_CHECKBOX, _SETCUSTOMCONTROL_TEXTBOX, _SETCUSTOMCONTROL_SELECTBOX, _SETCUSTOMCONTROL_BUTTON, _SETCUSTOMCONTROL_COLORPICKER, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='Control', full_name='Server.SetCustomControl.Control',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=2608,
  serialized_end=3467,
)

_SERVERCOMMAND.fields_by_name['add_custom_control'].message_type = _CUSTOMCONTROL
_SERVERCOMMAND.fields_by_name['set_camera'].message_type = _SETCAMERA
_SERVERCOMMAND.fields_by_name['add_object'].message_type = _ADDOBJECT
_SERVERCOMMAND.fields_by_name['set_key_event_handler'].message_type = _SETKEYEVENTHANDLER
_SERVERCOMMAND.fields_by_name['remove_object'].message_type = _REMOVEOBJECT
_SERVERCOMMAND.fields_by_name['remove_custom_control'].message_type = _REMOVECUSTOMCONTROL
_SERVERCOMMAND.fields_by_name['set_custom_control'].message_type = _SETCUSTOMCONTROL
_SERVERCOMMAND.oneofs_by_name['Command'].fields.append(
  _SERVERCOMMAND.fields_by_name['log_message'])
_SERVERCOMMAND.fields_by_name['log_message'].containing_oneof = _SERVERCOMMAND.oneofs_by_name['Command']
_SERVERCOMMAND.oneofs_by_name['Command'].fields.append(
  _SERVERCOMMAND.fields_by_name['capture_screen'])
_SERVERCOMMAND.fields_by_name['capture_screen'].containing_oneof = _SERVERCOMMAND.oneofs_by_name['Command']
_SERVERCOMMAND.oneofs_by_name['Command'].fields.append(
  _SERVERCOMMAND.fields_by_name['add_custom_control'])
_SERVERCOMMAND.fields_by_name['add_custom_control'].containing_oneof = _SERVERCOMMAND.oneofs_by_name['Command']
_SERVERCOMMAND.oneofs_by_name['Command'].fields.append(
  _SERVERCOMMAND.fields_by_name['set_camera'])
_SERVERCOMMAND.fields_by_name['set_camera'].containing_oneof = _SERVERCOMMAND.oneofs_by_name['Command']
_SERVERCOMMAND.oneofs_by_name['Command'].fields.append(
  _SERVERCOMMAND.fields_by_name['add_object'])
_SERVERCOMMAND.fields_by_name['add_object'].containing_oneof = _SERVERCOMMAND.oneofs_by_name['Command']
_SERVERCOMMAND.oneofs_by_name['Command'].fields.append(
  _SERVERCOMMAND.fields_by_name['set_key_event_handler'])
_SERVERCOMMAND.fields_by_name['set_key_event_handler'].containing_oneof = _SERVERCOMMAND.oneofs_by_name['Command']
_SERVERCOMMAND.oneofs_by_name['Command'].fields.append(
  _SERVERCOMMAND.fields_by_name['remove_object'])
_SERVERCOMMAND.fields_by_name['remove_object'].containing_oneof = _SERVERCOMMAND.oneofs_by_name['Command']
_SERVERCOMMAND.oneofs_by_name['Command'].fields.append(
  _SERVERCOMMAND.fields_by_name['remove_custom_control'])
_SERVERCOMMAND.fields_by_name['remove_custom_control'].containing_oneof = _SERVERCOMMAND.oneofs_by_name['Command']
_SERVERCOMMAND.oneofs_by_name['Command'].fields.append(
  _SERVERCOMMAND.fields_by_name['set_custom_control'])
_SERVERCOMMAND.fields_by_name['set_custom_control'].containing_oneof = _SERVERCOMMAND.oneofs_by_name['Command']
_CUSTOMCONTROL_SLIDER.containing_type = _CUSTOMCONTROL
_CUSTOMCONTROL_CHECKBOX.containing_type = _CUSTOMCONTROL
_CUSTOMCONTROL_TEXTBOX.containing_type = _CUSTOMCONTROL
_CUSTOMCONTROL_SELECTBOX.containing_type = _CUSTOMCONTROL
_CUSTOMCONTROL_BUTTON.containing_type = _CUSTOMCONTROL
_CUSTOMCONTROL_COLORPICKER.containing_type = _CUSTOMCONTROL
_CUSTOMCONTROL_FOLDER.containing_type = _CUSTOMCONTROL
_CUSTOMCONTROL.fields_by_name['slider'].message_type = _CUSTOMCONTROL_SLIDER
_CUSTOMCONTROL.fields_by_name['checkbox'].message_type = _CUSTOMCONTROL_CHECKBOX
_CUSTOMCONTROL.fields_by_name['textbox'].message_type = _CUSTOMCONTROL_TEXTBOX
_CUSTOMCONTROL.fields_by_name['selectbox'].message_type = _CUSTOMCONTROL_SELECTBOX
_CUSTOMCONTROL.fields_by_name['button'].message_type = _CUSTOMCONTROL_BUTTON
_CUSTOMCONTROL.fields_by_name['color_picker'].message_type = _CUSTOMCONTROL_COLORPICKER
_CUSTOMCONTROL.fields_by_name['folder'].message_type = _CUSTOMCONTROL_FOLDER
_CUSTOMCONTROL.oneofs_by_name['Control'].fields.append(
  _CUSTOMCONTROL.fields_by_name['slider'])
_CUSTOMCONTROL.fields_by_name['slider'].containing_oneof = _CUSTOMCONTROL.oneofs_by_name['Control']
_CUSTOMCONTROL.oneofs_by_name['Control'].fields.append(
  _CUSTOMCONTROL.fields_by_name['checkbox'])
_CUSTOMCONTROL.fields_by_name['checkbox'].containing_oneof = _CUSTOMCONTROL.oneofs_by_name['Control']
_CUSTOMCONTROL.oneofs_by_name['Control'].fields.append(
  _CUSTOMCONTROL.fields_by_name['textbox'])
_CUSTOMCONTROL.fields_by_name['textbox'].containing_oneof = _CUSTOMCONTROL.oneofs_by_name['Control']
_CUSTOMCONTROL.oneofs_by_name['Control'].fields.append(
  _CUSTOMCONTROL.fields_by_name['selectbox'])
_CUSTOMCONTROL.fields_by_name['selectbox'].containing_oneof = _CUSTOMCONTROL.oneofs_by_name['Control']
_CUSTOMCONTROL.oneofs_by_name['Control'].fields.append(
  _CUSTOMCONTROL.fields_by_name['button'])
_CUSTOMCONTROL.fields_by_name['button'].containing_oneof = _CUSTOMCONTROL.oneofs_by_name['Control']
_CUSTOMCONTROL.oneofs_by_name['Control'].fields.append(
  _CUSTOMCONTROL.fields_by_name['color_picker'])
_CUSTOMCONTROL.fields_by_name['color_picker'].containing_oneof = _CUSTOMCONTROL.oneofs_by_name['Control']
_CUSTOMCONTROL.oneofs_by_name['Control'].fields.append(
  _CUSTOMCONTROL.fields_by_name['folder'])
_CUSTOMCONTROL.fields_by_name['folder'].containing_oneof = _CUSTOMCONTROL.oneofs_by_name['Control']
_REMOVECUSTOMCONTROL.oneofs_by_name['Object'].fields.append(
  _REMOVECUSTOMCONTROL.fields_by_name['all'])
_REMOVECUSTOMCONTROL.fields_by_name['all'].containing_oneof = _REMOVECUSTOMCONTROL.oneofs_by_name['Object']
_REMOVECUSTOMCONTROL.oneofs_by_name['Object'].fields.append(
  _REMOVECUSTOMCONTROL.fields_by_name['by_uuid'])
_REMOVECUSTOMCONTROL.fields_by_name['by_uuid'].containing_oneof = _REMOVECUSTOMCONTROL.oneofs_by_name['Object']
_SETCAMERA.fields_by_name['position'].message_type = _VECXYZF
_SETCAMERA.fields_by_name['target'].message_type = _VECXYZF
_SETCAMERA.oneofs_by_name['Camera'].fields.append(
  _SETCAMERA.fields_by_name['perspective_fov'])
_SETCAMERA.fields_by_name['perspective_fov'].containing_oneof = _SETCAMERA.oneofs_by_name['Camera']
_SETCAMERA.oneofs_by_name['Camera'].fields.append(
  _SETCAMERA.fields_by_name['orthographic_frustum_height'])
_SETCAMERA.fields_by_name['orthographic_frustum_height'].containing_oneof = _SETCAMERA.oneofs_by_name['Camera']
_SETCAMERA.oneofs_by_name['Camera'].fields.append(
  _SETCAMERA.fields_by_name['position'])
_SETCAMERA.fields_by_name['position'].containing_oneof = _SETCAMERA.oneofs_by_name['Camera']
_SETCAMERA.oneofs_by_name['Camera'].fields.append(
  _SETCAMERA.fields_by_name['target'])
_SETCAMERA.fields_by_name['target'].containing_oneof = _SETCAMERA.oneofs_by_name['Camera']
_ADDOBJECT_LINESET.fields_by_name['points'].message_type = _VECXYZF
_ADDOBJECT_LINESET.fields_by_name['colors'].message_type = _VECRGBF
_ADDOBJECT_LINESET.containing_type = _ADDOBJECT
_ADDOBJECT_POINTCLOUD.containing_type = _ADDOBJECT
_ADDOBJECT_OVERLAY_IMAGE.containing_type = _ADDOBJECT_OVERLAY
_ADDOBJECT_OVERLAY.fields_by_name['position'].message_type = _VECXYZF
_ADDOBJECT_OVERLAY.fields_by_name['image'].message_type = _ADDOBJECT_OVERLAY_IMAGE
_ADDOBJECT_OVERLAY.fields_by_name['type'].enum_type = _ADDOBJECT_OVERLAY_COORDINATETYPE
_ADDOBJECT_OVERLAY.containing_type = _ADDOBJECT
_ADDOBJECT_OVERLAY_COORDINATETYPE.containing_type = _ADDOBJECT_OVERLAY
_ADDOBJECT_OVERLAY.oneofs_by_name['Contents'].fields.append(
  _ADDOBJECT_OVERLAY.fields_by_name['html'])
_ADDOBJECT_OVERLAY.fields_by_name['html'].containing_oneof = _ADDOBJECT_OVERLAY.oneofs_by_name['Contents']
_ADDOBJECT_OVERLAY.oneofs_by_name['Contents'].fields.append(
  _ADDOBJECT_OVERLAY.fields_by_name['image'])
_ADDOBJECT_OVERLAY.fields_by_name['image'].containing_oneof = _ADDOBJECT_OVERLAY.oneofs_by_name['Contents']
_ADDOBJECT_MESH.fields_by_name['points'].message_type = _VECXYZF
_ADDOBJECT_MESH.fields_by_name['colors'].message_type = _VECRGBF
_ADDOBJECT_MESH.containing_type = _ADDOBJECT
_ADDOBJECT.fields_by_name['line_set'].message_type = _ADDOBJECT_LINESET
_ADDOBJECT.fields_by_name['point_cloud'].message_type = _ADDOBJECT_POINTCLOUD
_ADDOBJECT.fields_by_name['overlay'].message_type = _ADDOBJECT_OVERLAY
_ADDOBJECT.fields_by_name['mesh'].message_type = _ADDOBJECT_MESH
_ADDOBJECT.oneofs_by_name['Object'].fields.append(
  _ADDOBJECT.fields_by_name['line_set'])
_ADDOBJECT.fields_by_name['line_set'].containing_oneof = _ADDOBJECT.oneofs_by_name['Object']
_ADDOBJECT.oneofs_by_name['Object'].fields.append(
  _ADDOBJECT.fields_by_name['point_cloud'])
_ADDOBJECT.fields_by_name['point_cloud'].containing_oneof = _ADDOBJECT.oneofs_by_name['Object']
_ADDOBJECT.oneofs_by_name['Object'].fields.append(
  _ADDOBJECT.fields_by_name['overlay'])
_ADDOBJECT.fields_by_name['overlay'].containing_oneof = _ADDOBJECT.oneofs_by_name['Object']
_ADDOBJECT.oneofs_by_name['Object'].fields.append(
  _ADDOBJECT.fields_by_name['mesh'])
_ADDOBJECT.fields_by_name['mesh'].containing_oneof = _ADDOBJECT.oneofs_by_name['Object']
_REMOVEOBJECT.oneofs_by_name['Object'].fields.append(
  _REMOVEOBJECT.fields_by_name['all'])
_REMOVEOBJECT.fields_by_name['all'].containing_oneof = _REMOVEOBJECT.oneofs_by_name['Object']
_REMOVEOBJECT.oneofs_by_name['Object'].fields.append(
  _REMOVEOBJECT.fields_by_name['by_uuid'])
_REMOVEOBJECT.fields_by_name['by_uuid'].containing_oneof = _REMOVEOBJECT.oneofs_by_name['Object']
_SETKEYEVENTHANDLER.oneofs_by_name['Event'].fields.append(
  _SETKEYEVENTHANDLER.fields_by_name['keydown'])
_SETKEYEVENTHANDLER.fields_by_name['keydown'].containing_oneof = _SETKEYEVENTHANDLER.oneofs_by_name['Event']
_SETKEYEVENTHANDLER.oneofs_by_name['Event'].fields.append(
  _SETKEYEVENTHANDLER.fields_by_name['keyup'])
_SETKEYEVENTHANDLER.fields_by_name['keyup'].containing_oneof = _SETKEYEVENTHANDLER.oneofs_by_name['Event']
_SETKEYEVENTHANDLER.oneofs_by_name['Event'].fields.append(
  _SETKEYEVENTHANDLER.fields_by_name['keypress'])
_SETKEYEVENTHANDLER.fields_by_name['keypress'].containing_oneof = _SETKEYEVENTHANDLER.oneofs_by_name['Event']
_SETCUSTOMCONTROL_SLIDER.containing_type = _SETCUSTOMCONTROL
_SETCUSTOMCONTROL_SLIDER.oneofs_by_name['_name'].fields.append(
  _SETCUSTOMCONTROL_SLIDER.fields_by_name['name'])
_SETCUSTOMCONTROL_SLIDER.fields_by_name['name'].containing_oneof = _SETCUSTOMCONTROL_SLIDER.oneofs_by_name['_name']
_SETCUSTOMCONTROL_SLIDER.oneofs_by_name['_min'].fields.append(
  _SETCUSTOMCONTROL_SLIDER.fields_by_name['min'])
_SETCUSTOMCONTROL_SLIDER.fields_by_name['min'].containing_oneof = _SETCUSTOMCONTROL_SLIDER.oneofs_by_name['_min']
_SETCUSTOMCONTROL_SLIDER.oneofs_by_name['_max'].fields.append(
  _SETCUSTOMCONTROL_SLIDER.fields_by_name['max'])
_SETCUSTOMCONTROL_SLIDER.fields_by_name['max'].containing_oneof = _SETCUSTOMCONTROL_SLIDER.oneofs_by_name['_max']
_SETCUSTOMCONTROL_SLIDER.oneofs_by_name['_step'].fields.append(
  _SETCUSTOMCONTROL_SLIDER.fields_by_name['step'])
_SETCUSTOMCONTROL_SLIDER.fields_by_name['step'].containing_oneof = _SETCUSTOMCONTROL_SLIDER.oneofs_by_name['_step']
_SETCUSTOMCONTROL_SLIDER.oneofs_by_name['_value'].fields.append(
  _SETCUSTOMCONTROL_SLIDER.fields_by_name['value'])
_SETCUSTOMCONTROL_SLIDER.fields_by_name['value'].containing_oneof = _SETCUSTOMCONTROL_SLIDER.oneofs_by_name['_value']
_SETCUSTOMCONTROL_CHECKBOX.containing_type = _SETCUSTOMCONTROL
_SETCUSTOMCONTROL_CHECKBOX.oneofs_by_name['_name'].fields.append(
  _SETCUSTOMCONTROL_CHECKBOX.fields_by_name['name'])
_SETCUSTOMCONTROL_CHECKBOX.fields_by_name['name'].containing_oneof = _SETCUSTOMCONTROL_CHECKBOX.oneofs_by_name['_name']
_SETCUSTOMCONTROL_CHECKBOX.oneofs_by_name['_value'].fields.append(
  _SETCUSTOMCONTROL_CHECKBOX.fields_by_name['value'])
_SETCUSTOMCONTROL_CHECKBOX.fields_by_name['value'].containing_oneof = _SETCUSTOMCONTROL_CHECKBOX.oneofs_by_name['_value']
_SETCUSTOMCONTROL_TEXTBOX.containing_type = _SETCUSTOMCONTROL
_SETCUSTOMCONTROL_TEXTBOX.oneofs_by_name['_name'].fields.append(
  _SETCUSTOMCONTROL_TEXTBOX.fields_by_name['name'])
_SETCUSTOMCONTROL_TEXTBOX.fields_by_name['name'].containing_oneof = _SETCUSTOMCONTROL_TEXTBOX.oneofs_by_name['_name']
_SETCUSTOMCONTROL_TEXTBOX.oneofs_by_name['_value'].fields.append(
  _SETCUSTOMCONTROL_TEXTBOX.fields_by_name['value'])
_SETCUSTOMCONTROL_TEXTBOX.fields_by_name['value'].containing_oneof = _SETCUSTOMCONTROL_TEXTBOX.oneofs_by_name['_value']
_SETCUSTOMCONTROL_SELECTBOX.containing_type = _SETCUSTOMCONTROL
_SETCUSTOMCONTROL_SELECTBOX.oneofs_by_name['_name'].fields.append(
  _SETCUSTOMCONTROL_SELECTBOX.fields_by_name['name'])
_SETCUSTOMCONTROL_SELECTBOX.fields_by_name['name'].containing_oneof = _SETCUSTOMCONTROL_SELECTBOX.oneofs_by_name['_name']
_SETCUSTOMCONTROL_SELECTBOX.oneofs_by_name['_value'].fields.append(
  _SETCUSTOMCONTROL_SELECTBOX.fields_by_name['value'])
_SETCUSTOMCONTROL_SELECTBOX.fields_by_name['value'].containing_oneof = _SETCUSTOMCONTROL_SELECTBOX.oneofs_by_name['_value']
_SETCUSTOMCONTROL_BUTTON.containing_type = _SETCUSTOMCONTROL
_SETCUSTOMCONTROL_BUTTON.oneofs_by_name['_name'].fields.append(
  _SETCUSTOMCONTROL_BUTTON.fields_by_name['name'])
_SETCUSTOMCONTROL_BUTTON.fields_by_name['name'].containing_oneof = _SETCUSTOMCONTROL_BUTTON.oneofs_by_name['_name']
_SETCUSTOMCONTROL_COLORPICKER.containing_type = _SETCUSTOMCONTROL
_SETCUSTOMCONTROL_COLORPICKER.oneofs_by_name['_name'].fields.append(
  _SETCUSTOMCONTROL_COLORPICKER.fields_by_name['name'])
_SETCUSTOMCONTROL_COLORPICKER.fields_by_name['name'].containing_oneof = _SETCUSTOMCONTROL_COLORPICKER.oneofs_by_name['_name']
_SETCUSTOMCONTROL_COLORPICKER.oneofs_by_name['_value'].fields.append(
  _SETCUSTOMCONTROL_COLORPICKER.fields_by_name['value'])
_SETCUSTOMCONTROL_COLORPICKER.fields_by_name['value'].containing_oneof = _SETCUSTOMCONTROL_COLORPICKER.oneofs_by_name['_value']
_SETCUSTOMCONTROL.fields_by_name['slider'].message_type = _SETCUSTOMCONTROL_SLIDER
_SETCUSTOMCONTROL.fields_by_name['checkbox'].message_type = _SETCUSTOMCONTROL_CHECKBOX
_SETCUSTOMCONTROL.fields_by_name['textbox'].message_type = _SETCUSTOMCONTROL_TEXTBOX
_SETCUSTOMCONTROL.fields_by_name['selectbox'].message_type = _SETCUSTOMCONTROL_SELECTBOX
_SETCUSTOMCONTROL.fields_by_name['button'].message_type = _SETCUSTOMCONTROL_BUTTON
_SETCUSTOMCONTROL.fields_by_name['color_picker'].message_type = _SETCUSTOMCONTROL_COLORPICKER
_SETCUSTOMCONTROL.oneofs_by_name['Control'].fields.append(
  _SETCUSTOMCONTROL.fields_by_name['slider'])
_SETCUSTOMCONTROL.fields_by_name['slider'].containing_oneof = _SETCUSTOMCONTROL.oneofs_by_name['Control']
_SETCUSTOMCONTROL.oneofs_by_name['Control'].fields.append(
  _SETCUSTOMCONTROL.fields_by_name['checkbox'])
_SETCUSTOMCONTROL.fields_by_name['checkbox'].containing_oneof = _SETCUSTOMCONTROL.oneofs_by_name['Control']
_SETCUSTOMCONTROL.oneofs_by_name['Control'].fields.append(
  _SETCUSTOMCONTROL.fields_by_name['textbox'])
_SETCUSTOMCONTROL.fields_by_name['textbox'].containing_oneof = _SETCUSTOMCONTROL.oneofs_by_name['Control']
_SETCUSTOMCONTROL.oneofs_by_name['Control'].fields.append(
  _SETCUSTOMCONTROL.fields_by_name['selectbox'])
_SETCUSTOMCONTROL.fields_by_name['selectbox'].containing_oneof = _SETCUSTOMCONTROL.oneofs_by_name['Control']
_SETCUSTOMCONTROL.oneofs_by_name['Control'].fields.append(
  _SETCUSTOMCONTROL.fields_by_name['button'])
_SETCUSTOMCONTROL.fields_by_name['button'].containing_oneof = _SETCUSTOMCONTROL.oneofs_by_name['Control']
_SETCUSTOMCONTROL.oneofs_by_name['Control'].fields.append(
  _SETCUSTOMCONTROL.fields_by_name['color_picker'])
_SETCUSTOMCONTROL.fields_by_name['color_picker'].containing_oneof = _SETCUSTOMCONTROL.oneofs_by_name['Control']
DESCRIPTOR.message_types_by_name['ServerCommand'] = _SERVERCOMMAND
DESCRIPTOR.message_types_by_name['CustomControl'] = _CUSTOMCONTROL
DESCRIPTOR.message_types_by_name['RemoveCustomControl'] = _REMOVECUSTOMCONTROL
DESCRIPTOR.message_types_by_name['SetCamera'] = _SETCAMERA
DESCRIPTOR.message_types_by_name['AddObject'] = _ADDOBJECT
DESCRIPTOR.message_types_by_name['RemoveObject'] = _REMOVEOBJECT
DESCRIPTOR.message_types_by_name['SetKeyEventHandler'] = _SETKEYEVENTHANDLER
DESCRIPTOR.message_types_by_name['VecXYZf'] = _VECXYZF
DESCRIPTOR.message_types_by_name['VecRGBf'] = _VECRGBF
DESCRIPTOR.message_types_by_name['SetCustomControl'] = _SETCUSTOMCONTROL
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ServerCommand = _reflection.GeneratedProtocolMessageType('ServerCommand', (_message.Message,), {
  'DESCRIPTOR' : _SERVERCOMMAND,
  '__module__' : 'protobuf.server_pb2'
  # @@protoc_insertion_point(class_scope:Server.ServerCommand)
  })
_sym_db.RegisterMessage(ServerCommand)

CustomControl = _reflection.GeneratedProtocolMessageType('CustomControl', (_message.Message,), {

  'Slider' : _reflection.GeneratedProtocolMessageType('Slider', (_message.Message,), {
    'DESCRIPTOR' : _CUSTOMCONTROL_SLIDER,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.CustomControl.Slider)
    })
  ,

  'CheckBox' : _reflection.GeneratedProtocolMessageType('CheckBox', (_message.Message,), {
    'DESCRIPTOR' : _CUSTOMCONTROL_CHECKBOX,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.CustomControl.CheckBox)
    })
  ,

  'TextBox' : _reflection.GeneratedProtocolMessageType('TextBox', (_message.Message,), {
    'DESCRIPTOR' : _CUSTOMCONTROL_TEXTBOX,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.CustomControl.TextBox)
    })
  ,

  'SelectBox' : _reflection.GeneratedProtocolMessageType('SelectBox', (_message.Message,), {
    'DESCRIPTOR' : _CUSTOMCONTROL_SELECTBOX,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.CustomControl.SelectBox)
    })
  ,

  'Button' : _reflection.GeneratedProtocolMessageType('Button', (_message.Message,), {
    'DESCRIPTOR' : _CUSTOMCONTROL_BUTTON,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.CustomControl.Button)
    })
  ,

  'ColorPicker' : _reflection.GeneratedProtocolMessageType('ColorPicker', (_message.Message,), {
    'DESCRIPTOR' : _CUSTOMCONTROL_COLORPICKER,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.CustomControl.ColorPicker)
    })
  ,

  'Folder' : _reflection.GeneratedProtocolMessageType('Folder', (_message.Message,), {
    'DESCRIPTOR' : _CUSTOMCONTROL_FOLDER,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.CustomControl.Folder)
    })
  ,
  'DESCRIPTOR' : _CUSTOMCONTROL,
  '__module__' : 'protobuf.server_pb2'
  # @@protoc_insertion_point(class_scope:Server.CustomControl)
  })
_sym_db.RegisterMessage(CustomControl)
_sym_db.RegisterMessage(CustomControl.Slider)
_sym_db.RegisterMessage(CustomControl.CheckBox)
_sym_db.RegisterMessage(CustomControl.TextBox)
_sym_db.RegisterMessage(CustomControl.SelectBox)
_sym_db.RegisterMessage(CustomControl.Button)
_sym_db.RegisterMessage(CustomControl.ColorPicker)
_sym_db.RegisterMessage(CustomControl.Folder)

RemoveCustomControl = _reflection.GeneratedProtocolMessageType('RemoveCustomControl', (_message.Message,), {
  'DESCRIPTOR' : _REMOVECUSTOMCONTROL,
  '__module__' : 'protobuf.server_pb2'
  # @@protoc_insertion_point(class_scope:Server.RemoveCustomControl)
  })
_sym_db.RegisterMessage(RemoveCustomControl)

SetCamera = _reflection.GeneratedProtocolMessageType('SetCamera', (_message.Message,), {
  'DESCRIPTOR' : _SETCAMERA,
  '__module__' : 'protobuf.server_pb2'
  # @@protoc_insertion_point(class_scope:Server.SetCamera)
  })
_sym_db.RegisterMessage(SetCamera)

AddObject = _reflection.GeneratedProtocolMessageType('AddObject', (_message.Message,), {

  'LineSet' : _reflection.GeneratedProtocolMessageType('LineSet', (_message.Message,), {
    'DESCRIPTOR' : _ADDOBJECT_LINESET,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.AddObject.LineSet)
    })
  ,

  'PointCloud' : _reflection.GeneratedProtocolMessageType('PointCloud', (_message.Message,), {
    'DESCRIPTOR' : _ADDOBJECT_POINTCLOUD,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.AddObject.PointCloud)
    })
  ,

  'Overlay' : _reflection.GeneratedProtocolMessageType('Overlay', (_message.Message,), {

    'Image' : _reflection.GeneratedProtocolMessageType('Image', (_message.Message,), {
      'DESCRIPTOR' : _ADDOBJECT_OVERLAY_IMAGE,
      '__module__' : 'protobuf.server_pb2'
      # @@protoc_insertion_point(class_scope:Server.AddObject.Overlay.Image)
      })
    ,
    'DESCRIPTOR' : _ADDOBJECT_OVERLAY,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.AddObject.Overlay)
    })
  ,

  'Mesh' : _reflection.GeneratedProtocolMessageType('Mesh', (_message.Message,), {
    'DESCRIPTOR' : _ADDOBJECT_MESH,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.AddObject.Mesh)
    })
  ,
  'DESCRIPTOR' : _ADDOBJECT,
  '__module__' : 'protobuf.server_pb2'
  # @@protoc_insertion_point(class_scope:Server.AddObject)
  })
_sym_db.RegisterMessage(AddObject)
_sym_db.RegisterMessage(AddObject.LineSet)
_sym_db.RegisterMessage(AddObject.PointCloud)
_sym_db.RegisterMessage(AddObject.Overlay)
_sym_db.RegisterMessage(AddObject.Overlay.Image)
_sym_db.RegisterMessage(AddObject.Mesh)

RemoveObject = _reflection.GeneratedProtocolMessageType('RemoveObject', (_message.Message,), {
  'DESCRIPTOR' : _REMOVEOBJECT,
  '__module__' : 'protobuf.server_pb2'
  # @@protoc_insertion_point(class_scope:Server.RemoveObject)
  })
_sym_db.RegisterMessage(RemoveObject)

SetKeyEventHandler = _reflection.GeneratedProtocolMessageType('SetKeyEventHandler', (_message.Message,), {
  'DESCRIPTOR' : _SETKEYEVENTHANDLER,
  '__module__' : 'protobuf.server_pb2'
  # @@protoc_insertion_point(class_scope:Server.SetKeyEventHandler)
  })
_sym_db.RegisterMessage(SetKeyEventHandler)

VecXYZf = _reflection.GeneratedProtocolMessageType('VecXYZf', (_message.Message,), {
  'DESCRIPTOR' : _VECXYZF,
  '__module__' : 'protobuf.server_pb2'
  # @@protoc_insertion_point(class_scope:Server.VecXYZf)
  })
_sym_db.RegisterMessage(VecXYZf)

VecRGBf = _reflection.GeneratedProtocolMessageType('VecRGBf', (_message.Message,), {
  'DESCRIPTOR' : _VECRGBF,
  '__module__' : 'protobuf.server_pb2'
  # @@protoc_insertion_point(class_scope:Server.VecRGBf)
  })
_sym_db.RegisterMessage(VecRGBf)

SetCustomControl = _reflection.GeneratedProtocolMessageType('SetCustomControl', (_message.Message,), {

  'Slider' : _reflection.GeneratedProtocolMessageType('Slider', (_message.Message,), {
    'DESCRIPTOR' : _SETCUSTOMCONTROL_SLIDER,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.SetCustomControl.Slider)
    })
  ,

  'CheckBox' : _reflection.GeneratedProtocolMessageType('CheckBox', (_message.Message,), {
    'DESCRIPTOR' : _SETCUSTOMCONTROL_CHECKBOX,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.SetCustomControl.CheckBox)
    })
  ,

  'TextBox' : _reflection.GeneratedProtocolMessageType('TextBox', (_message.Message,), {
    'DESCRIPTOR' : _SETCUSTOMCONTROL_TEXTBOX,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.SetCustomControl.TextBox)
    })
  ,

  'SelectBox' : _reflection.GeneratedProtocolMessageType('SelectBox', (_message.Message,), {
    'DESCRIPTOR' : _SETCUSTOMCONTROL_SELECTBOX,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.SetCustomControl.SelectBox)
    })
  ,

  'Button' : _reflection.GeneratedProtocolMessageType('Button', (_message.Message,), {
    'DESCRIPTOR' : _SETCUSTOMCONTROL_BUTTON,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.SetCustomControl.Button)
    })
  ,

  'ColorPicker' : _reflection.GeneratedProtocolMessageType('ColorPicker', (_message.Message,), {
    'DESCRIPTOR' : _SETCUSTOMCONTROL_COLORPICKER,
    '__module__' : 'protobuf.server_pb2'
    # @@protoc_insertion_point(class_scope:Server.SetCustomControl.ColorPicker)
    })
  ,
  'DESCRIPTOR' : _SETCUSTOMCONTROL,
  '__module__' : 'protobuf.server_pb2'
  # @@protoc_insertion_point(class_scope:Server.SetCustomControl)
  })
_sym_db.RegisterMessage(SetCustomControl)
_sym_db.RegisterMessage(SetCustomControl.Slider)
_sym_db.RegisterMessage(SetCustomControl.CheckBox)
_sym_db.RegisterMessage(SetCustomControl.TextBox)
_sym_db.RegisterMessage(SetCustomControl.SelectBox)
_sym_db.RegisterMessage(SetCustomControl.Button)
_sym_db.RegisterMessage(SetCustomControl.ColorPicker)


# @@protoc_insertion_point(module_scope)



import numpy

import numpy.random

from cumo.pointcloudviewer import DownSampleStrategy


def down_sample_pointcloud(pc: numpy.ndarray, strategy: DownSampleStrategy, max_num_points: int) -> numpy.ndarray:
    assert pc.shape[1] == 3 or pc.shape[1] == 4

    if pc.shape[0] <= max_num_points:
        return pc

    if strategy == DownSampleStrategy.NONE:
        return pc
    if strategy == DownSampleStrategy.RANDOM_SAMPLE:
        return down_sample_random(pc, max_num_points)
    if strategy == DownSampleStrategy.VOXEL_GRID:
        assert strategy.voxel_size is not None
        return down_sample_voxel(pc, strategy.voxel_size, max_num_points)

    assert False, f"strategy not implemented: {strategy.name}"


def down_sample_random(pc: numpy.ndarray, max_num_points: int) -> numpy.ndarray:
    indices = numpy.arange(pc.shape[0])
    return pc[numpy.random.choice(indices, max_num_points)]


def down_sample_voxel(pc: numpy.ndarray, voxel_size: float, max_num_points: int) -> numpy.ndarray:
    scale = 1.0 / voxel_size
    voxels = {}
    output = []

    for i, _ in enumerate(pc):
        [x, y, z] = numpy.round(pc[i][:3] * scale).astype(numpy.int32)
        (p, c, n) = voxels.get((x, y, z), ((0.0, 0.0, 0.0), pc[i][3], 0))
        voxels[(x, y, z)] = (pc[i][:3] + p, c, n + 1)

    for ((x, y, z), (acc, c, n)) in voxels.items():
        output.append(numpy.hstack((acc / n, c)))

    output = numpy.array(output).astype(numpy.float32)

    if len(output) > max_num_points:
        return down_sample_random(output, max_num_points)
    return output
